
From: Ian Kent <raven@themaw.net>

It's possible for an event wait request to arive before the event
requestor.  If this happens the daemon never gets notified and autofs
hangs.

Signed-off-by: Ian Kent <raven@themaw.net>
Signed-off-by: Andrew Morton <akpm@osdl.org>
---

 25-akpm/fs/autofs4/autofs_i.h |    1 +
 25-akpm/fs/autofs4/waitq.c    |   22 ++++++++++++----------
 2 files changed, 13 insertions(+), 10 deletions(-)

diff -puN fs/autofs4/autofs_i.h~autofs4-wait-order-fix fs/autofs4/autofs_i.h
--- 25/fs/autofs4/autofs_i.h~autofs4-wait-order-fix	2005-04-10 15:25:40.000000000 -0700
+++ 25-akpm/fs/autofs4/autofs_i.h	2005-04-10 15:25:40.000000000 -0700
@@ -84,6 +84,7 @@ struct autofs_wait_queue {
 	char *name;
 	/* This is for status reporting upon return */
 	int status;
+	atomic_t notified;
 	atomic_t wait_ctr;
 };
 
diff -puN fs/autofs4/waitq.c~autofs4-wait-order-fix fs/autofs4/waitq.c
--- 25/fs/autofs4/waitq.c~autofs4-wait-order-fix	2005-04-10 15:25:40.000000000 -0700
+++ 25-akpm/fs/autofs4/waitq.c	2005-04-10 15:25:40.000000000 -0700
@@ -210,17 +210,8 @@ int autofs4_wait(struct autofs_sb_info *
 		wq->len = len;
 		wq->status = -EINTR; /* Status return if interrupted */
 		atomic_set(&wq->wait_ctr, 2);
+		atomic_set(&wq->notified, 1);
 		up(&sbi->wq_sem);
-
-		DPRINTK("new wait id = 0x%08lx, name = %.*s, nfy=%d",
-			(unsigned long) wq->wait_queue_token, wq->len, wq->name, notify);
-		/* autofs4_notify_daemon() may block */
-		if (notify != NFY_NONE) {
-			autofs4_notify_daemon(sbi,wq, 
-					notify == NFY_MOUNT ?
-						  autofs_ptype_missing :
-						  autofs_ptype_expire_multi);
-		}
 	} else {
 		atomic_inc(&wq->wait_ctr);
 		up(&sbi->wq_sem);
@@ -229,6 +220,17 @@ int autofs4_wait(struct autofs_sb_info *
 			(unsigned long) wq->wait_queue_token, wq->len, wq->name, notify);
 	}
 
+	if (notify != NFY_NONE && atomic_dec_and_test(&wq->notified)) {
+		int type = (notify == NFY_MOUNT ?
+			autofs_ptype_missing : autofs_ptype_expire_multi);
+
+		DPRINTK(("new wait id = 0x%08lx, name = %.*s, nfy=%d\n",
+			(unsigned long) wq->wait_queue_token, wq->len, wq->name, notify));
+
+		/* autofs4_notify_daemon() may block */
+		autofs4_notify_daemon(sbi, wq, type);
+	}
+
 	/* wq->name is NULL if and only if the lock is already released */
 
 	if ( sbi->catatonic ) {
_
